# Simplified Java virtual machine
# Steve Checkoway (steve@cs)
# 18 October 2004
#
#------------{ global data section } ------------#
# declare all global variables and string        #
# constants in this section.                     #
#------------------------------------------------#
.data
error:	.asciiz "The java program has performed an illegal instruction:\n"
badop:	.asciiz "Invalid opcode: "
badsk:	.asciiz "Tried to pop past the bottom of the stack.\n"
badsk2:	.asciiz "Tried to push too much.\n"
result:	.asciiz "Result: "
newln:	.asciiz "\n"
comma:	.asciiz ", "
# A sample set of bytecodes.
		.align  2
jvmbytecode:
		.byte   0x10, 0x7f						# bipush
		.byte   0x10, 0x22						# bipush
		.byte   0x60							# iadd
		.byte   0xac							# ireturn
		.byte   0x00							# nothing
		.align  2
end_jvmbytecode:

jvmstack: .space 1024 # Allocate stack space
jvmlocal: .word 0,0,0,0 # Reserve space for the 4 answers
		.word 12 # length of the array to be examined
		.word 13,5,7,2,13,234,13,19,20,17,23,14 #an example array
		.space 956
#------------{ code section }--------------------#
# place all main code and procedure code in      #
# this section of the file                       #
#------------------------------------------------#

.text

# This never returns, it only prints out an error message and then quits.
# a0 - 0 means print the string pointed to by a1, nonzero mean print the
# string pointed to by a1 and the in pointed to by a2
# a1 - the string's address
# a2 - the int to print
illegal:
		move $t0, $a0			# how much are we printing?
		move $t1, $a2			# we might need this
		li $v0, 4				# print string
		move $a0, $a1			# print_string wants the argument in a0
		syscall					# print it
		beq $t0, $0, die		# only print one thing => goto die
		li $v0, 1				# print int
		move $a0, $t1			# copy the int to print
		syscall					# print it
		li $v0, 4				# print string
		la $a0, newln			# new line
		syscall					# print it
die:	li $v0, 10				# exit
		syscall					# does not return
.end illegal


JVM:	subu $sp, $sp, 4		# store fp
		sw $fp, 4($sp)			# push fp
		move $fp, $sp			# set up the frame pointer

		# register usage
		# a0 jvm pc
		# a1 jvm sp
		# a2 jvm local storage pointer
		# t0 opcode
		# t1 jvm v1
		# t2 jvm v2
		# t3 bottom of the jvm stack
		# t4 temp
		# t5 temp

		move $t3, $a1			# set the bottom of the stack

		# First load and decode the opcode
top:	lbu $t0, 0($a0)			# load opcode
		beq $t0, 0x60, iadd		# if (op == IADD) goto iadd
		beq $t0, 0x64, isub		# if (op == ISUB) goto isub
		beq $t0, 0x68, imul		# ... goto imul
		beq $t0, 0x6b, idiv		# ... ... idiv
		beq $t0, 0x15, iload	# ... iload
		beq $t0, 0x36, istore	# istore
		beq $t0, 0x2e, iaload	# iaload
		beq $t0, 0x4f, iastore	# iastore
		beq $t0, 0x10, bipush	# bipush
		beq $t0, 0x17, sipush	# sipush
		beq $t0, 0x59, dup		# dup
		beq $t0, 0x5c, dup2		# dup2
		beq $t0, 0x57, pop		# pop
		beq $t0, 0x99, ifeq		# ifeq
		beq $t0, 0x9a, ifne		# ifne
		beq $t0, 0x9b, iflt		# iflt
		beq $t0, 0x9c, ifge		# ifge
		beq $t0, 0x9d, ifgt		# ifgt
		beq $t0, 0x9e, ifle		# ifle
		beq $t0, 0xa7, goto		# goto
		beq $t0, 0xac, ireturn	# ireturn
		beqz $t0, none			# void return
	
		move $a2, $a0			# copy the opcode
		li $a0, 1				# print from both a1 and a2
		la $a1, badop			# bad opcode
		j illegal				# illegal instruction

iadd:	subu $t4, $a1, 8		# two operands get popped
		bge $t4, $t3, L1		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L1:		lw $t1, -4($a1)			# pop v1
		lw $t2, -8($a1)			# pop v2
		add $t1, $t2, $t1		# v1 = v2 + v1
		sw $t1, -8($a1)			# push v1
		subu $a1, $a1, 4		# take one word away from the stack
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

isub:	subu $t4, $a1, 8		# two operands get popped
		bge $t4, $t3, L2		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L2:		lw $t1, -4($a1)			# pop v1
		lw $t2, -8($a1)			# pop v2
		sub $t1, $t2, $t1		# v1 = v2 - v1
		sw $t1, -8($a1)			# push v1
		subu $a1, $a1, 4		# take one word away from the stack
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

imul:	subu $t4, $a1, 8		# two operands get popped
		bge $t4, $t3, L3		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L3:		lw $t1, -4($a1)			# pop v1
		lw $t2, -8($a1)			# pop v2
		mul $t1, $t2, $t1		# v1 = v2 * v1
		sw $t1, -8($a1)			# push v1
		subu $a1, $a1, 4		# take one word away from the stack
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

idiv:	subu $t4, $a1, 8		# two operands get popped
		bge $t4, $t3, L4		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L4:		lw $t1, -4($a1)			# pop v1
		lw $t2, -8($a1)			# pop v2
		div $t1, $t2, $t1		# v1 = v2 / v1
		sw $t1, -8($a1)			# push v1
		subu $a1, $a1, 4		# take one word away from the stack
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

# The instructions say that we don't have to worry about local overflow here
# Still, the stack could grow too large
iload:	blt $a1, $a2, L5		# check that the stack is < the local storage
		li $a0, 0				# only a1
		la $a1, badsk2			# bad stack 2 (push)
		j illegal				# illegal instruction
L5:		lbu $t4, 1($a0)			# load index
		sll $t4, $t4, 2			# multiply by 4 to get a byte offset
		add $t4, $a2, $t4		# get the address of the word
		lw $t1, 0($t4)			# load the word into v1
		sw $t1, 0($a1)			# push v1
		addiu $a1, $a1, 4		# add word to the stack
		addiu $a0, $a0, 2		# increment the jvm's pc
		b top					# continue

istore:	bgt $a1, $t3, L6		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L6:		lw $t1, -4($a1)			# pop v1
		subu $a1, $a1, 4		# one word popped
		lbu $t4, 1($a0)			# load index
		sll $t4, $t4, 2			# multiply by 4 to get a byte offset
		add $t4, $a2, $t4		# get the address
		sw $t1, 0($t4)			# store the local word
		addiu $a0, $a0, 2		# increment the jvm's pc
		b top					# continue

iaload:	bgt $a1, $t3, L7		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L7:		lw $t1, -4($a1)			# pop v1
		sll $t4, $t1, 2			# multiply by 4 to get a byte offset
		add $t4, $a2, $t4		# get the address
		lw $t1, 0($t4)			# load the local word into v1
		sw $t1, -4($a1)			# push v1
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

iastore:
		subu $t4, $a1, 8		# two operands get popped
		bge $t4, $t3, L8		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L8:		lw $t1, -4($a1)			# pop v1
		lw $t2, -8($a1)			# pop v2
		sll $t4, $t1, 2			# multiply by 4 to get a byte offset
		add $t4, $a2, $t4		# get address
		sw $t2, 0($t4)			# store v2 at local address
		subu $a1, $a1, 8		# remove two words
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

bipush:	blt $a1, $a2, L9		# check that the stack is < the local storage
		li $a0, 0				# only a1
		la $a1, badsk2			# bad stack 2 (push)
		j illegal				# illegal instruction
L9:	lb $t4, 1($a0)				# load the byte and sign extends
		sw $t4, 0($a1)			# push the byte
		addiu $a1, $a1, 4		# add one word
		addiu $a0, $a0, 2		# increment the jvm's pc
		b top					# continue

sipush:	blt $a1, $a2, L10		# check that the stack is < the local storage
		li $a0, 0				# only a1
		la $a1, badsk2			# bad stack 2 (push)
		j illegal				# illegal instruction
L10:	lb $t4, 1($a0)			# load first byte
		lbu $t5, 2($a0)			# load the second byte, unsigned
		sll $t4, $t4, 8			# temp1 <<= 8
		or $t4, $t4, $t5		# temp1 |= temp2
		sw $t4, 0($a1)			# push the short int
		addiu $a1, $a1, 4		# add one word
		addiu $a0, $a0, 3		# increment the jvm's pc
		b top					# continue
		
# Need to check that dup and dup2 don't underflow as well as overflow
dup:	blt $a1, $a2, L11		# check that the stack is < the local storage
		li $a0, 0				# only a1
		la $a1, badsk2			# bad stack 2 (push)
		j illegal				# illegal instruction
L11:	bgt $a1, $t3, L11_2		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal
L11_2:	lw $t1, -4($a1)			# copy v1
		sw $t1, 0($a1)			# push v1
		addiu $a1, $a1, 4		# add one word
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

dup2:	subu $t4, $a2, 4		# must be able to push 2 words on
		blt $a1, $t4, L12		# if there is room, keep working
		li $a0, 0				# only a1
		la $a1, badsk2			# bad stack 2 (push)
		j illegal				# illegal instruction
L12:	subu $t4, $a1, 8		# two operands get popped
		bge $t4, $t3, L12_2		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L12_2:	lw $t1, -4($a1)			# copy v1
		lw $t2, -8($a1)			# copy v2
		sw $t2, 0($a1)			# push v2
		sw $t1, 4($a1)			# push v1
		addiu $a1, $a1, 8		# add two words
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

pop:	bgt $a1, $t3, L13		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L13:	subu $a1, $a1, 4		# pop it
		addiu $a0, $a0, 1		# increment the jvm's pc
		b top					# continue

ifeq:	bgt $a1, $t3, L14		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L14:	lw $t1, -4($a1)			# pop v1
		subu $a1, $a1, 4		# pop it
		beq $t1, $0, goto		# if (v1 == 0) goto goto
		addiu $a0, $a0, 3		# increment pc
		b top					# continue
		
ifne:	bgt $a1, $t3, L15		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L15:	lw $t1, -4($a1)			# pop v1
		subu $a1, $a1, 4		# pop it
		bne $t1, $0, goto		# if (v1 == 0) goto goto
		addiu $a0, $a0, 3		# increment pc
		b top					# continue

iflt:	bgt $a1, $t3, L16		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L16:	lw $t1, -4($a1)			# pop v1
		subu $a1, $a1, 4		# pop it
		blt $t1, $0, goto		# if (v1 == 0) goto goto
		addiu $a0, $a0, 3		# increment pc
		b top					# continue

ifge:	bgt $a1, $t3, L17		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L17:	lw $t1, -4($a1)			# pop v1
		subu $a1, $a1, 4		# pop it
		bge $t1, $0, goto		# if (v1 == 0) goto goto
		addiu $a0, $a0, 3		# increment pc
		b top					# continue

ifgt:	bgt $a1, $t3, L18		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L18:	lw $t1, -4($a1)			# pop v1
		subu $a1, $a1, 4		# pop it
		bgt $t1, $0, goto		# if (v1 == 0) goto goto
		addiu $a0, $a0, 3		# increment pc
		b top					# continue

ifle:	bgt $a1, $t3, L19		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L19:	lw $t1, -4($a1)			# pop v1
		subu $a1, $a1, 4		# pop it
		ble $t1, $0, goto		# if (v1 == 0) goto goto
		addiu $a0, $a0, 3		# increment pc
		b top					# continue
		
goto:	lb $t4, 1($a0)			# load first offset
		lbu $t5, 2($a0)			# load second offset
		sll $t4, $t4, 8			# t4 <<= 8
		or $t4, $t4, $t5		# t4 |= t5
		add $a0, $a0, $t4		# branch
		b top					# continue
		
ireturn:
		bgt $a1, $t3, L20		# if there is room, continue on
		li $a0, 0				# only a1
		la $a1, badsk			# bad stack
		j illegal				# illegal instruction
L20:	lw $v0, -4($a1)			# pop v1 into return reg
		li $v1, 0				# valid return value
		subu $a1, $a1, 4		# really pop it off
		b done					# goto done
	
none:	li $v1, 1				# no return value
		
		#clean up
done:	lw $fp, 4($sp)			# pop fp
		addiu $sp, $sp, 4		# fix up the stack pointer
		jr $ra					# return
.end JVM
	
.globl main
main:	subu $sp, $sp, 4		# create 1 new word on the stack
		sw $ra, 4($sp)			# store the return address
		#call the JVM interpreter after storing its arguments in the
		#appropriate registers
		la $a0, jvmbytecode	#store the bytecode address in $a0
		la $a1, jvmstack		# store the stack address in $a1
		la $a2, jvmlocal		# store the local address in $a2
		jal JVM					# call the Java Virtual Machine
		
		# Print out the result
		move $t0, $v0			# copy the value
		move $t1, $v1			# copy the return value.
		li $v0, 4				# print string
		la $a0, result			# print result string
		syscall					# print it
		bnez $t1, noret			# goto noret if there is no return value.
		li $v0, 1				# print int
		move $a0, $t0			# print the return value
		syscall					# print it

		#Print out the first four memory address as {1,2,3,4} 
noret:	li $v0, 11				# print char
		li $a0, ' '
		syscall
		li $v0, 11
		li $a0, '{'
		syscall
		la $t0, jvmlocal
		li $v0, 1
		lw $a0, 0($t0)
		syscall
		li $v0, 11
		li $a0, ','
		syscall
		li $v0, 1
		lw $a0, 4($t0)
		syscall
		li $v0, 11
		li $a0, ','
		syscall
		li $v0, 1
		lw $a0, 8($t0)
		syscall
		li $v0, 11
		li $a0, ','
		syscall
		li $v0, 1
		lw $a0, 12($t0)
		syscall
		li $v0, 11
		li $a0, '}'
		syscall
		
		li $v0, 4				# print new line
		la $a0, newln			# print new line
		syscall					# print it
	
		# When the JVM is done, we have nothing to do, so just exit our
		# program.
		lw $ra, 4($sp)			# restore the return address
		addiu $sp, $sp, 4		# restore the stack
		jr $ra					# return
.end main
